package java.util.concurrent;

import java.lang.invoke.SerializedLambda;
import java.util.function.Consumer;
import java.util.Collections;
import java.util.AbstractCollection;
import java.util.SortedSet;
import java.util.Spliterator;
import java.util.AbstractSet;
import java.util.NavigableMap;
import java.util.function.BiConsumer;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.Collection;
import java.util.NavigableSet;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.SortedMap;
import java.util.Map;
import sun.misc.Unsafe;
import java.util.Comparator;
import java.io.Serializable;
import java.util.AbstractMap;

public class ConcurrentSkipListMap<K, V> extends AbstractMap<K, V> implements ConcurrentNavigableMap<K, V>, Cloneable, Serializable
{
    private static final long serialVersionUID = -8627078645895051609L;
    private static final Object BASE_HEADER;
    private transient volatile ConcurrentSkipListMap.HeadIndex<K, V> head;
    final Comparator<? super K> comparator;
    private transient ConcurrentSkipListMap.KeySet<K> keySet;
    private transient ConcurrentSkipListMap.EntrySet<K, V> entrySet;
    private transient ConcurrentSkipListMap.Values<V> values;
    private transient ConcurrentNavigableMap<K, V> descendingMap;
    private static final int EQ = 1;
    private static final int LT = 2;
    private static final int GT = 0;
    private static final Unsafe UNSAFE;
    private static final long headOffset;
    private static final long SECONDARY;
    
    private void initialize() {
        this.keySet = null;
        this.entrySet = null;
        this.values = null;
        this.descendingMap = null;
        this.head = new ConcurrentSkipListMap.HeadIndex(new ConcurrentSkipListMap.Node(null, ConcurrentSkipListMap.BASE_HEADER), null, null, 1);
    }
    
    private boolean casHead(ConcurrentSkipListMap.HeadIndex<K, V> expected, ConcurrentSkipListMap.HeadIndex<K, V> x) {
        return ConcurrentSkipListMap.UNSAFE.compareAndSwapObject(this, ConcurrentSkipListMap.headOffset, expected, x);
    }
    
    static final int cpr(Comparator comparator, Object o, Object o2) {
        return (comparator != null) ? comparator.compare(o, o2) : ((Comparable)o).compareTo(o2);
    }
    
    private ConcurrentSkipListMap.Node<K, V> findPredecessor(Object o, Comparator<? super K> comparator) {
        if (o == null) {
            throw new NullPointerException();
        }
        ConcurrentSkipListMap.HeadIndex<Object, Object> head;
    Block_6:
        while (true) {
            head = this.head;
            ConcurrentSkipListMap.Index<Object, Object> index = head.right;
            while (true) {
                if (index != null) {
                    ConcurrentSkipListMap.Node<Object, Object> node = index.node;
                    Object key = node.key;
                    if (node.value == null) {
                        if (!head.unlink(index)) {
                            break;
                        }
                        index = head.right;
                        continue;
                    }
                    else if (cpr(comparator, o, key) > 0) {
                        head = index;
                        index = index.right;
                        continue;
                    }
                }
                ConcurrentSkipListMap.Index<Object, Object> down;
                if ((down = head.down) == null) {
                    break Block_6;
                }
                head = down;
                index = down.right;
            }
        }
        return head.node;
    }
    
    private ConcurrentSkipListMap.Node<K, V> findNode(Object o) {
        if (o == null) {
            throw new NullPointerException();
        }
        Comparator<Object> comparator = this.comparator;
    Label_0017:
        while (true) {
            ConcurrentSkipListMap.Node<K, V> predecessor = this.findPredecessor(o, comparator);
            ConcurrentSkipListMap.Node<Object, Object> next = predecessor.next;
            while (next != null) {
                ConcurrentSkipListMap.Node<Object, Object> next2 = next.next;
                if (next == predecessor.next) {
                    Object value;
                    if ((value = next.value) == null) {
                        next.helpDelete(predecessor, next2);
                    }
                    else if (predecessor.value != null && value != next) {
                        int cpr;
                        if ((cpr = cpr(comparator, o, next.key)) == 0) {
                            return next;
                        }
                        if (cpr < 0) {
                            break;
                        }
                        predecessor = next;
                        next = next2;
                        continue;
                    }
                }
                continue Label_0017;
            }
            break;
        }
        return null;
    }
    
    private V doGet(Object o) {
        if (o == null) {
            throw new NullPointerException();
        }
        Comparator<Object> comparator = this.comparator;
    Label_0017:
        while (true) {
            ConcurrentSkipListMap.Node<K, V> predecessor = this.findPredecessor(o, comparator);
            ConcurrentSkipListMap.Node<Object, Object> next = predecessor.next;
            while (next != null) {
                ConcurrentSkipListMap.Node<Object, Object> next2 = next.next;
                if (next == predecessor.next) {
                    Object value;
                    if ((value = next.value) == null) {
                        next.helpDelete(predecessor, next2);
                    }
                    else if (predecessor.value != null && value != next) {
                        int cpr;
                        if ((cpr = cpr(comparator, o, next.key)) == 0) {
                            return value;
                        }
                        if (cpr < 0) {
                            break;
                        }
                        predecessor = next;
                        next = next2;
                        continue;
                    }
                }
                continue Label_0017;
            }
            break;
        }
        return null;
    }
    
    private V doPut(K k, V v, boolean b) {
        if (k == null) {
            throw new NullPointerException();
        }
        Comparator<Object> comparator = this.comparator;
    Label_0018:
        while (true) {
            ConcurrentSkipListMap.Node<K, V> predecessor = this.findPredecessor(k, comparator);
            ConcurrentSkipListMap.Node<Object, Object> next = predecessor.next;
            Label_0183: {
                while (next != null) {
                    ConcurrentSkipListMap.Node<Object, Object> next2 = next.next;
                    if (next == predecessor.next) {
                        Object value;
                        if ((value = next.value) == null) {
                            next.helpDelete(predecessor, next2);
                        }
                        else if (predecessor.value != null && value != next) {
                            int cpr;
                            if ((cpr = cpr(comparator, k, next.key)) > 0) {
                                predecessor = next;
                                next = next2;
                                continue;
                            }
                            if (cpr == 0) {
                                if (b || next.casValue(value, v)) {
                                    return value;
                                }
                            }
                            else {
                                break;
                            }
                        }
                    }
                    continue Label_0018;
                }
                ConcurrentSkipListMap.Node o = new ConcurrentSkipListMap.Node<Object, Object>(k, v, next);
                if (!predecessor.casNext(next, o)) {
                    continue;
                }
                int nextSecondarySeed = ThreadLocalRandom.nextSecondarySeed();
                if ((nextSecondarySeed & 0x80000001) == 0x0) {
                    int n = 1;
                    while (((nextSecondarySeed >>>= 1) & 0x1) != 0x0) {
                        ++n;
                    }
                    ConcurrentSkipListMap.Index down = null;
                    ConcurrentSkipListMap.HeadIndex headIndex = this.head;
                    int level;
                    if (n <= (level = headIndex.level)) {
                        for (int l = 1; l <= n; ++l) {
                            down = new ConcurrentSkipListMap.Index(o, down);
                        }
                    }
                    else {
                        n = level + 1;
                        ConcurrentSkipListMap.Index[] array = (ConcurrentSkipListMap.Index[])new ConcurrentSkipListMap.Index[n + 1];
                        for (int i = 1; i <= n; ++i) {
                            down = (array[i] = new ConcurrentSkipListMap.Index(o, down));
                        }
                        while (true) {
                            headIndex = this.head;
                            int level2 = headIndex.level;
                            if (n <= level2) {
                                break;
                            }
                            ConcurrentSkipListMap.Node<Object, Object> node = headIndex.node;
                            for (int j = level2 + 1; j <= n; ++j) {
                                headIndex = new ConcurrentSkipListMap.HeadIndex(node, headIndex, array[j], j);
                            }
                            if (this.casHead(headIndex, headIndex)) {
                                down = array[n = level2];
                                break;
                            }
                            continue;
                        }
                    }
                Label_0437:
                    while (true) {
                        int level3 = headIndex.level;
                        ConcurrentSkipListMap.Index<Object, Object> index = headIndex.right;
                        while (headIndex != null) {
                            if (down == null) {
                                break;
                            }
                            if (index != null) {
                                ConcurrentSkipListMap.Node<Object, Object> node2 = index.node;
                                int cpr2 = cpr(comparator, k, node2.key);
                                if (node2.value == null) {
                                    if (!headIndex.unlink(index)) {
                                        continue Label_0641;
                                    }
                                    index = headIndex.right;
                                    continue;
                                }
                                else if (cpr2 > 0) {
                                    headIndex = index;
                                    index = index.right;
                                    continue;
                                }
                            }
                            if (level3 == n) {
                                if (!headIndex.link(index, down)) {
                                    continue Label_0641;
                                }
                                if (down.node.value == null) {
                                    this.findNode(k);
                                    break;
                                }
                                if (--n == 0) {
                                    break;
                                }
                            }
                            if (--level3 >= n && level3 < n) {
                                down = down.down;
                            }
                            headIndex = headIndex.down;
                            index = headIndex.right;
                            continue;
                            Label_0641: {
                                continue Label_0437;
                            }
                        }
                        break;
                    }
                }
                return null;
            }
        }
    }
    
    final V doRemove(Object o, Object o2) {
        if (o == null) {
            throw new NullPointerException();
        }
        Comparator<Object> comparator = this.comparator;
    Label_0017:
        while (true) {
            ConcurrentSkipListMap.Node<K, V> predecessor = this.findPredecessor(o, comparator);
            ConcurrentSkipListMap.Node<Object, Object> next = predecessor.next;
            while (next != null) {
                ConcurrentSkipListMap.Node<Object, Object> next2 = next.next;
                if (next == predecessor.next) {
                    Object value;
                    if ((value = next.value) == null) {
                        next.helpDelete(predecessor, next2);
                    }
                    else if (predecessor.value != null && value != next) {
                        int cpr;
                        if ((cpr = cpr(comparator, o, next.key)) < 0) {
                            break;
                        }
                        if (cpr > 0) {
                            predecessor = next;
                            next = next2;
                            continue;
                        }
                        if (o2 != null && !o2.equals(value)) {
                            break;
                        }
                        if (next.casValue(value, null)) {
                            if (!next.appendMarker(next2) || !predecessor.casNext(next, next2)) {
                                this.findNode(o);
                            }
                            else {
                                this.findPredecessor(o, comparator);
                                if (this.head.right == null) {
                                    this.tryReduceLevel();
                                }
                            }
                            return value;
                        }
                    }
                }
                continue Label_0017;
            }
            break;
        }
        return null;
    }
    
    private void tryReduceLevel() {
        ConcurrentSkipListMap.HeadIndex<K, V> head = this.head;
        ConcurrentSkipListMap.HeadIndex<K, V> headIndex;
        ConcurrentSkipListMap.HeadIndex<Object, Object> headIndex2;
        if (head.level > 3 && (headIndex = (ConcurrentSkipListMap.HeadIndex)head.down) != null && (headIndex2 = (ConcurrentSkipListMap.HeadIndex)headIndex.down) != null && headIndex2.right == null && headIndex.right == null && head.right == null && this.casHead(head, headIndex) && head.right != null) {
            this.casHead(headIndex, head);
        }
    }
    
    final ConcurrentSkipListMap.Node<K, V> findFirst() {
        ConcurrentSkipListMap.Node<K, V> node;
        ConcurrentSkipListMap.Node<Object, Object> next;
        while ((next = (node = this.head.node).next) != null) {
            if (next.value != null) {
                return next;
            }
            next.helpDelete(node, next.next);
        }
        return null;
    }
    
    private Map.Entry<K, V> doRemoveFirstEntry() {
        ConcurrentSkipListMap.Node<Object, Object> node;
        ConcurrentSkipListMap.Node<Object, Object> next;
        while ((next = (node = this.head.node).next) != null) {
            ConcurrentSkipListMap.Node<Object, Object> next2 = next.next;
            if (next != node.next) {
                continue;
            }
            Object value = next.value;
            if (value == null) {
                next.helpDelete(node, next2);
            }
            else {
                if (!next.casValue(value, null)) {
                    continue;
                }
                if (!next.appendMarker(next2) || !node.casNext(next, next2)) {
                    this.findFirst();
                }
                this.clearIndexToFirst();
                return new AbstractMap.SimpleImmutableEntry(next.key, value);
            }
        }
        return null;
    }
    
    private void clearIndexToFirst() {
    Block_3:
        while (true) {
            ConcurrentSkipListMap.HeadIndex<Object, Object> headIndex = this.head;
            while (true) {
                ConcurrentSkipListMap.Index<Object, Object> right = headIndex.right;
                if (right != null && right.indexesDeletedNode() && !headIndex.unlink(right)) {
                    break;
                }
                if ((headIndex = headIndex.down) == null) {
                    break Block_3;
                }
                continue;
            }
        }
        if (this.head.right == null) {
            this.tryReduceLevel();
        }
    }
    
    private Map.Entry<K, V> doRemoveLastEntry() {
        while (true) {
            ConcurrentSkipListMap.Node<K, V> predecessorOfLast = this.findPredecessorOfLast();
            ConcurrentSkipListMap.Node<Object, Object> next = predecessorOfLast.next;
            if (next == null) {
                if (!predecessorOfLast.isBaseHeader()) {
                    continue;
                }
                return null;
            }
            else {
                while (true) {
                    ConcurrentSkipListMap.Node<Object, Object> next2 = next.next;
                    if (next != predecessorOfLast.next) {
                        break;
                    }
                    Object value = next.value;
                    if (value == null) {
                        next.helpDelete(predecessorOfLast, next2);
                        break;
                    }
                    if (predecessorOfLast.value != null) {
                        if (value == next) {
                            break;
                        }
                        if (next2 != null) {
                            predecessorOfLast = next;
                            next = next2;
                        }
                        else {
                            if (!next.casValue(value, null)) {
                                break;
                            }
                            Object key = next.key;
                            if (!next.appendMarker(next2) || !predecessorOfLast.casNext(next, next2)) {
                                this.findNode(key);
                            }
                            else {
                                this.findPredecessor(key, this.comparator);
                                if (this.head.right == null) {
                                    this.tryReduceLevel();
                                }
                            }
                            return new AbstractMap.SimpleImmutableEntry(key, value);
                        }
                    }
                    else {
                        break;
                    }
                }
            }
        }
    }
    
    final ConcurrentSkipListMap.Node<K, V> findLast() {
        ConcurrentSkipListMap.HeadIndex<Object, Object> headIndex = this.head;
        ConcurrentSkipListMap.Node<Object, Object> node;
        while (true) {
            ConcurrentSkipListMap.Index<Object, Object> right;
            if ((right = headIndex.right) != null) {
                if (right.indexesDeletedNode()) {
                    headIndex.unlink(right);
                    headIndex = this.head;
                }
                else {
                    headIndex = right;
                }
            }
            else {
                ConcurrentSkipListMap.Index<Object, Object> down;
                if ((down = headIndex.down) != null) {
                    headIndex = down;
                }
                else {
                    node = headIndex.node;
                    ConcurrentSkipListMap.Node<Object, Object> next = node.next;
                    while (next != null) {
                        ConcurrentSkipListMap.Node<Object, Object> next2 = next.next;
                        if (next == node.next) {
                            Object value = next.value;
                            if (value == null) {
                                next.helpDelete(node, next2);
                            }
                            else if (node.value != null && value != next) {
                                node = next;
                                next = next2;
                                continue;
                            }
                        }
                        headIndex = this.head;
                        continue Label_0165;
                    }
                    break;
                }
            }
            Label_0165:;
        }
        return node.isBaseHeader() ? null : node;
    }
    
    private ConcurrentSkipListMap.Node<K, V> findPredecessorOfLast() {
        ConcurrentSkipListMap.HeadIndex<Object, Object> head;
    Label_0059:
        while (true) {
            head = this.head;
            while (true) {
                ConcurrentSkipListMap.Index<Object, Object> right;
                if ((right = head.right) != null) {
                    if (right.indexesDeletedNode()) {
                        head.unlink(right);
                        break;
                    }
                    if (right.node.next != null) {
                        head = right;
                        continue;
                    }
                }
                ConcurrentSkipListMap.Index<Object, Object> down;
                if ((down = head.down) != null) {
                    head = down;
                }
                else {
                    break Label_0059;
                }
            }
        }
        return head.node;
    }
    
    final ConcurrentSkipListMap.Node<K, V> findNear(K k, int n, Comparator<? super K> comparator) {
        if (k == null) {
            throw new NullPointerException();
        }
    Label_0012:
        while (true) {
            ConcurrentSkipListMap.Node<K, V> predecessor = this.findPredecessor(k, comparator);
            ConcurrentSkipListMap.Node<Object, Object> next = predecessor.next;
            while (next != null) {
                ConcurrentSkipListMap.Node<Object, Object> next2 = next.next;
                if (next == predecessor.next) {
                    Object value;
                    if ((value = next.value) == null) {
                        next.helpDelete(predecessor, next2);
                    }
                    else if (predecessor.value != null && value != next) {
                        int cpr = cpr(comparator, k, next.key);
                        if ((cpr == 0 && (n & 0x1) != 0x0) || (cpr < 0 && (n & 0x2) == 0x0)) {
                            return next;
                        }
                        if (cpr <= 0 && (n & 0x2) != 0x0) {
                            return predecessor.isBaseHeader() ? null : predecessor;
                        }
                        predecessor = next;
                        next = next2;
                        continue;
                    }
                }
                continue Label_0012;
            }
            return ((n & 0x2) == 0x0 || predecessor.isBaseHeader()) ? null : predecessor;
        }
    }
    
    final AbstractMap.SimpleImmutableEntry<K, V> getNear(K k, int n) {
        Comparator<Object> comparator = this.comparator;
        while (true) {
            ConcurrentSkipListMap.Node<K, V> near = this.findNear(k, n, comparator);
            if (near == null) {
                return null;
            }
            AbstractMap.SimpleImmutableEntry<Object, Object> snapshot = near.createSnapshot();
            if (snapshot != null) {
                return snapshot;
            }
            continue;
        }
    }
    
    public ConcurrentSkipListMap() {
        this.comparator = null;
        this.initialize();
    }
    
    public ConcurrentSkipListMap(Comparator<? super K> comparator) {
        this.comparator = comparator;
        this.initialize();
    }
    
    public ConcurrentSkipListMap(Map<? extends K, ? extends V> m) {
        this.comparator = null;
        this.initialize();
        this.putAll(m);
    }
    
    public ConcurrentSkipListMap(SortedMap<K, ? extends V> sortedMap) {
        this.comparator = sortedMap.comparator();
        this.initialize();
        this.buildFromSorted(sortedMap);
    }
    
    public ConcurrentSkipListMap<K, V> clone() {
        try {
            ConcurrentSkipListMap<Object, Object> concurrentSkipListMap = (ConcurrentSkipListMap)super.clone();
            concurrentSkipListMap.initialize();
            concurrentSkipListMap.buildFromSorted(this);
            return concurrentSkipListMap;
        }
        catch (CloneNotSupportedException ex) {
            throw new InternalError();
        }
    }
    
    private void buildFromSorted(SortedMap<K, ? extends V> sortedMap) {
        if (sortedMap == null) {
            throw new NullPointerException();
        }
        ConcurrentSkipListMap.HeadIndex headIndex = this.head;
        ConcurrentSkipListMap.Node<Object, Object> node = headIndex.node;
        ArrayList list = new ArrayList();
        for (int i = 0; i <= headIndex.level; ++i) {
            list.add(null);
        }
        for (int j = headIndex.level; j > 0; headIndex = headIndex.down, --j) {
            list.set(j, headIndex);
        }
        for (Map.Entry entry : sortedMap.entrySet()) {
            int nextInt = ThreadLocalRandom.current().nextInt();
            int n = 0;
            if ((nextInt & 0x80000001) == 0x0) {
                do {
                    ++n;
                } while (((nextInt >>>= 1) & 0x1) != 0x0);
                if (n > headIndex.level) {
                    n = headIndex.level + 1;
                }
            }
            Object key = entry.getKey();
            Object value = entry.getValue();
            if (key == null || value == null) {
                throw new NullPointerException();
            }
            ConcurrentSkipListMap.Node next = new ConcurrentSkipListMap.Node<Object, Object>(key, value);
            node.next = next;
            node = next;
            if (n > 0) {
                Object o = null;
                for (int k = 1; k <= n; ++k) {
                    o = new ConcurrentSkipListMap.Index<Object, Object>(next, o);
                    if (k > headIndex.level) {
                        headIndex = new ConcurrentSkipListMap.HeadIndex(headIndex.node, headIndex, o, k);
                    }
                    if (k < list.size()) {
                        list.set(k, ((ConcurrentSkipListMap.Index)list.get(k)).right = o);
                    }
                    else {
                        list.add(o);
                    }
                }
            }
            else {
                continue;
            }
        }
        this.head = headIndex;
        return;
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        objectOutputStream.defaultWriteObject();
        for (ConcurrentSkipListMap.Node node = this.findFirst(); node != null; node = node.next) {
            Object validValue = node.getValidValue();
            if (validValue != null) {
                objectOutputStream.writeObject(node.key);
                objectOutputStream.writeObject(validValue);
            }
        }
        objectOutputStream.writeObject(null);
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        objectInputStream.defaultReadObject();
        this.initialize();
        ConcurrentSkipListMap.HeadIndex headIndex = this.head;
        ConcurrentSkipListMap.Node<Object, Object> node = headIndex.node;
        ArrayList list = new ArrayList();
        for (int i = 0; i <= headIndex.level; ++i) {
            list.add(null);
        }
        for (int j = headIndex.level; j > 0; headIndex = headIndex.down, --j) {
            list.set(j, headIndex);
        }
        while (true) {
            Object object = objectInputStream.readObject();
            if (object == null) {
                this.head = headIndex;
                return;
            }
            Object object2 = objectInputStream.readObject();
            if (object2 == null) {
                throw new NullPointerException();
            }
            int nextInt = ThreadLocalRandom.current().nextInt();
            int n = 0;
            if ((nextInt & 0x80000001) == 0x0) {
                do {
                    ++n;
                } while (((nextInt >>>= 1) & 0x1) != 0x0);
                if (n > headIndex.level) {
                    n = headIndex.level + 1;
                }
            }
            ConcurrentSkipListMap.Node next = new ConcurrentSkipListMap.Node<Object, Object>(object, object2);
            node.next = next;
            node = next;
            if (n > 0) {
                Object o = null;
                for (int k = 1; k <= n; ++k) {
                    o = new ConcurrentSkipListMap.Index<Object, Object>(next, o);
                    if (k > headIndex.level) {
                        headIndex = new ConcurrentSkipListMap.HeadIndex(headIndex.node, headIndex, o, k);
                    }
                    if (k < list.size()) {
                        list.set(k, ((ConcurrentSkipListMap.Index)list.get(k)).right = o);
                    }
                    else {
                        list.add(o);
                    }
                }
            }
            else {
                continue;
            }
        }
    }
    
    @Override
    public boolean containsKey(Object o) {
        return this.doGet(o) != null;
    }
    
    @Override
    public V get(Object o) {
        return this.doGet(o);
    }
    
    @Override
    public V getOrDefault(Object o, V v) {
        Object doGet;
        return ((doGet = this.doGet(o)) == null) ? v : doGet;
    }
    
    @Override
    public V put(K k, V v) {
        if (v == null) {
            throw new NullPointerException();
        }
        return this.doPut(k, v, false);
    }
    
    @Override
    public V remove(Object o) {
        return this.doRemove(o, null);
    }
    
    @Override
    public boolean containsValue(Object o) {
        if (o == null) {
            throw new NullPointerException();
        }
        ConcurrentSkipListMap.Node<K, V> node = this.findFirst();
        while (node != null) {
            Object validValue = node.getValidValue();
            if (validValue != null && o.equals(validValue)) {
                return true;
            }
            node = node.next;
        }
        return false;
    }
    
    @Override
    public int size() {
        long n = 0L;
        for (ConcurrentSkipListMap.Node node = this.findFirst(); node != null; node = node.next) {
            if (node.getValidValue() != null) {
                ++n;
            }
        }
        return (n >= 2147483647L) ? Integer.MAX_VALUE : ((int)n);
    }
    
    @Override
    public boolean isEmpty() {
        return this.findFirst() == null;
    }
    
    @Override
    public void clear() {
        while (true) {
            ConcurrentSkipListMap.HeadIndex<K, V> head = this.head;
            ConcurrentSkipListMap.HeadIndex<K, V> headIndex = (ConcurrentSkipListMap.HeadIndex)head.down;
            if (headIndex != null) {
                this.casHead(head, headIndex);
            }
            else {
                ConcurrentSkipListMap.Node<Object, Object> node;
                ConcurrentSkipListMap.Node<Object, Object> next;
                if ((node = head.node) != null && (next = node.next) != null) {
                    ConcurrentSkipListMap.Node<Object, Object> next2 = next.next;
                    if (next == node.next) {
                        Object value = next.value;
                        if (value == null) {
                            next.helpDelete(node, next2);
                        }
                        else if (next.casValue(value, null) && next.appendMarker(next2)) {
                            node.casNext(next, next2);
                        }
                        else {
                            continue;
                        }
                    }
                    else {
                        continue;
                    }
                }
                else {
                    break;
                }
            }
        }
    }
    
    @Override
    public V computeIfAbsent(K k, Function<? super K, ? extends V> function) {
        if (k == null || function == null) {
            throw new NullPointerException();
        }
        Object doGet;
        Object apply;
        if ((doGet = this.doGet(k)) == null && (apply = function.apply(k)) != null) {
            Object doPut;
            doGet = (((doPut = this.doPut(k, apply, true)) == null) ? apply : doPut);
            return doGet;
        }
        return doGet;
    }
    
    @Override
    public V computeIfPresent(K k, BiFunction<? super K, ? super V, ? extends V> biFunction) {
        if (k == null || biFunction == null) {
            throw new NullPointerException();
        }
        ConcurrentSkipListMap.Node<K, V> node;
        while ((node = this.findNode(k)) != null) {
            Object value;
            if ((value = node.value) == null) {
                continue;
            }
            Object apply = biFunction.apply(k, value);
            if (apply != null) {
                if (node.casValue(value, apply)) {
                    return apply;
                }
                continue;
            }
            else {
                if (this.doRemove(k, value) != null) {
                    break;
                }
                continue;
            }
        }
        return null;
    }
    
    @Override
    public V compute(K k, BiFunction<? super K, ? super V, ? extends V> biFunction) {
        if (k == null || biFunction == null) {
            throw new NullPointerException();
        }
        while (true) {
            ConcurrentSkipListMap.Node<K, V> node;
            if ((node = this.findNode(k)) == null) {
                Object apply2;
                if ((apply2 = biFunction.apply(k, null)) == null) {
                    break;
                }
                if (this.doPut(k, apply2, true) == null) {
                    return apply2;
                }
                continue;
            }
            else {
                Object value;
                if ((value = node.value) != null) {
                    Object apply;
                    if ((apply = biFunction.apply(k, value)) != null) {
                        if (node.casValue(value, apply)) {
                            return apply;
                        }
                        continue;
                    }
                    else {
                        if (this.doRemove(k, value) != null) {
                            break;
                        }
                        continue;
                    }
                }
                else {
                    continue;
                }
            }
        }
        return null;
    }
    
    @Override
    public V merge(K k, V v, BiFunction<? super V, ? super V, ? extends V> biFunction) {
        if (k == null || v == null || biFunction == null) {
            throw new NullPointerException();
        }
        while (true) {
            ConcurrentSkipListMap.Node<K, V> node;
            if ((node = this.findNode(k)) == null) {
                if (this.doPut(k, v, true) == null) {
                    return v;
                }
                continue;
            }
            else {
                Object value;
                if ((value = node.value) != null) {
                    Object apply;
                    if ((apply = biFunction.apply(value, v)) != null) {
                        if (node.casValue(value, apply)) {
                            return apply;
                        }
                        continue;
                    }
                    else {
                        if (this.doRemove(k, value) != null) {
                            return null;
                        }
                        continue;
                    }
                }
                else {
                    continue;
                }
            }
        }
    }
    
    @Override
    public NavigableSet<K> keySet() {
        ConcurrentSkipListMap.KeySet<Object> keySet = this.keySet;
        return (keySet != null) ? keySet : (this.keySet = new ConcurrentSkipListMap.KeySet(this));
    }
    
    @Override
    public NavigableSet<K> navigableKeySet() {
        ConcurrentSkipListMap.KeySet<Object> keySet = this.keySet;
        return (keySet != null) ? keySet : (this.keySet = new ConcurrentSkipListMap.KeySet(this));
    }
    
    @Override
    public Collection<V> values() {
        ConcurrentSkipListMap.Values<Object> values = this.values;
        return (values != null) ? values : (this.values = new ConcurrentSkipListMap.Values(this));
    }
    
    @Override
    public Set<Map.Entry<K, V>> entrySet() {
        ConcurrentSkipListMap.EntrySet<Object, Object> entrySet = this.entrySet;
        return (entrySet != null) ? entrySet : (this.entrySet = new ConcurrentSkipListMap.EntrySet(this));
    }
    
    @Override
    public ConcurrentNavigableMap<K, V> descendingMap() {
        ConcurrentNavigableMap<Object, Object> descendingMap = this.descendingMap;
        return (descendingMap != null) ? descendingMap : (this.descendingMap = new ConcurrentSkipListMap.SubMap(this, null, false, null, false, true));
    }
    
    @Override
    public NavigableSet<K> descendingKeySet() {
        return this.descendingMap().navigableKeySet();
    }
    
    @Override
    public boolean equals(Object o) {
        if (o == this) {
            return true;
        }
        if (!(o instanceof Map)) {
            return false;
        }
        Map<Object, Object> map = (Map)o;
        try {
            for (Map.Entry entry : this.entrySet()) {
                if (!entry.getValue().equals(map.get(entry.getKey()))) {
                    return false;
                }
                continue;
            }
            for (Map.Entry entry2 : map.entrySet()) {
                Object key = entry2.getKey();
                Object value = entry2.getValue();
                if (key == null || value == null || !value.equals(this.get(key))) {
                    return false;
                }
                continue;
            }
            return true;
        }
        catch (ClassCastException ex) {
            return false;
        }
        catch (NullPointerException ex2) {
            return false;
        }
    }
    
    @Override
    public V putIfAbsent(K k, V v) {
        if (v == null) {
            throw new NullPointerException();
        }
        return this.doPut(k, v, true);
    }
    
    @Override
    public boolean remove(Object o, Object o2) {
        if (o == null) {
            throw new NullPointerException();
        }
        return o2 != null && this.doRemove(o, o2) != null;
    }
    
    @Override
    public boolean replace(K k, V v, V v2) {
        if (k == null || v == null || v2 == null) {
            throw new NullPointerException();
        }
        ConcurrentSkipListMap.Node<K, V> node;
        while ((node = this.findNode(k)) != null) {
            Object value;
            if ((value = node.value) != null) {
                if (!v.equals(value)) {
                    return false;
                }
                if (node.casValue(value, v2)) {
                    return true;
                }
                continue;
            }
            else {
                continue;
            }
        }
        return false;
    }
    
    @Override
    public V replace(K k, V v) {
        if (k == null || v == null) {
            throw new NullPointerException();
        }
        ConcurrentSkipListMap.Node<K, V> node;
        while ((node = this.findNode(k)) != null) {
            Object value;
            if ((value = node.value) != null && node.casValue(value, v)) {
                return value;
            }
            continue;
        }
        return null;
    }
    
    @Override
    public Comparator<? super K> comparator() {
        return this.comparator;
    }
    
    @Override
    public K firstKey() {
        ConcurrentSkipListMap.Node<K, V> first = this.findFirst();
        if (first == null) {
            throw new NoSuchElementException();
        }
        return first.key;
    }
    
    @Override
    public K lastKey() {
        ConcurrentSkipListMap.Node<K, V> last = this.findLast();
        if (last == null) {
            throw new NoSuchElementException();
        }
        return last.key;
    }
    
    @Override
    public ConcurrentNavigableMap<K, V> subMap(K k, boolean b, K i, boolean b2) {
        if (k == null || i == null) {
            throw new NullPointerException();
        }
        return new ConcurrentSkipListMap.SubMap(this, k, b, i, b2, false);
    }
    
    @Override
    public ConcurrentNavigableMap<K, V> headMap(K k, boolean b) {
        if (k == null) {
            throw new NullPointerException();
        }
        return new ConcurrentSkipListMap.SubMap(this, null, false, k, b, false);
    }
    
    @Override
    public ConcurrentNavigableMap<K, V> tailMap(K k, boolean b) {
        if (k == null) {
            throw new NullPointerException();
        }
        return new ConcurrentSkipListMap.SubMap(this, k, b, null, false, false);
    }
    
    @Override
    public ConcurrentNavigableMap<K, V> subMap(K k, K i) {
        return this.subMap(k, true, i, false);
    }
    
    @Override
    public ConcurrentNavigableMap<K, V> headMap(K k) {
        return this.headMap(k, false);
    }
    
    @Override
    public ConcurrentNavigableMap<K, V> tailMap(K k) {
        return this.tailMap(k, true);
    }
    
    @Override
    public Map.Entry<K, V> lowerEntry(K k) {
        return this.getNear(k, 2);
    }
    
    @Override
    public K lowerKey(K k) {
        ConcurrentSkipListMap.Node<K, V> near = this.findNear(k, 2, this.comparator);
        return (near == null) ? null : near.key;
    }
    
    @Override
    public Map.Entry<K, V> floorEntry(K k) {
        return this.getNear(k, 3);
    }
    
    @Override
    public K floorKey(K k) {
        ConcurrentSkipListMap.Node<K, V> near = this.findNear(k, 3, this.comparator);
        return (near == null) ? null : near.key;
    }
    
    @Override
    public Map.Entry<K, V> ceilingEntry(K k) {
        return this.getNear(k, 1);
    }
    
    @Override
    public K ceilingKey(K k) {
        ConcurrentSkipListMap.Node<K, V> near = this.findNear(k, 1, this.comparator);
        return (near == null) ? null : near.key;
    }
    
    @Override
    public Map.Entry<K, V> higherEntry(K k) {
        return this.getNear(k, 0);
    }
    
    @Override
    public K higherKey(K k) {
        ConcurrentSkipListMap.Node<K, V> near = this.findNear(k, 0, this.comparator);
        return (near == null) ? null : near.key;
    }
    
    @Override
    public Map.Entry<K, V> firstEntry() {
        while (true) {
            ConcurrentSkipListMap.Node<K, V> first = this.findFirst();
            if (first == null) {
                return null;
            }
            AbstractMap.SimpleImmutableEntry<Object, Object> snapshot = first.createSnapshot();
            if (snapshot != null) {
                return snapshot;
            }
            continue;
        }
    }
    
    @Override
    public Map.Entry<K, V> lastEntry() {
        while (true) {
            ConcurrentSkipListMap.Node<K, V> last = this.findLast();
            if (last == null) {
                return null;
            }
            AbstractMap.SimpleImmutableEntry<Object, Object> snapshot = last.createSnapshot();
            if (snapshot != null) {
                return snapshot;
            }
            continue;
        }
    }
    
    @Override
    public Map.Entry<K, V> pollFirstEntry() {
        return this.doRemoveFirstEntry();
    }
    
    @Override
    public Map.Entry<K, V> pollLastEntry() {
        return this.doRemoveLastEntry();
    }
    
    Iterator<K> keyIterator() {
        return new ConcurrentSkipListMap.KeyIterator();
    }
    
    Iterator<V> valueIterator() {
        return new ConcurrentSkipListMap.ValueIterator();
    }
    
    Iterator<Map.Entry<K, V>> entryIterator() {
        return new ConcurrentSkipListMap.EntryIterator();
    }
    
    static final <E> List<E> toList(Collection<E> collection) {
        ArrayList<Object> list = new ArrayList();
        Iterator<Object> iterator = collection.iterator();
        while (iterator.hasNext()) {
            list.add(iterator.next());
        }
        return list;
    }
    
    @Override
    public void forEach(BiConsumer<? super K, ? super V> biConsumer) {
        if (biConsumer == null) {
            throw new NullPointerException();
        }
        for (ConcurrentSkipListMap.Node node = this.findFirst(); node != null; node = node.next) {
            Object validValue;
            if ((validValue = node.getValidValue()) != null) {
                biConsumer.accept(node.key, validValue);
            }
        }
        return;
    }
    
    @Override
    public void replaceAll(BiFunction<? super K, ? super V, ? extends V> biFunction) {
        if (biFunction == null) {
            throw new NullPointerException();
        }
        for (ConcurrentSkipListMap.Node node = this.findFirst(); node != null; node = node.next) {
            Object validValue;
            while ((validValue = node.getValidValue()) != null) {
                Object apply = biFunction.apply(node.key, validValue);
                if (apply == null) {
                    throw new NullPointerException();
                }
                if (node.casValue(validValue, apply)) {
                    break;
                }
                continue;
            }
        }
        return;
    }
    
    final ConcurrentSkipListMap.KeySpliterator<K, V> keySpliterator() {
        Comparator<Object> comparator = this.comparator;
        ConcurrentSkipListMap.HeadIndex<Object, Object> head;
        ConcurrentSkipListMap.Node<Object, Object> next;
        while (true) {
            ConcurrentSkipListMap.Node<K, V> node = (head = this.head).node;
            if ((next = node.next) == null || next.value != null) {
                break;
            }
            next.helpDelete(node, next.next);
        }
        return new ConcurrentSkipListMap.KeySpliterator(comparator, head, next, null, (next == null) ? 0 : Integer.MAX_VALUE);
    }
    
    final ConcurrentSkipListMap.ValueSpliterator<K, V> valueSpliterator() {
        Comparator<Object> comparator = this.comparator;
        ConcurrentSkipListMap.HeadIndex<Object, Object> head;
        ConcurrentSkipListMap.Node<Object, Object> next;
        while (true) {
            ConcurrentSkipListMap.Node<K, V> node = (head = this.head).node;
            if ((next = node.next) == null || next.value != null) {
                break;
            }
            next.helpDelete(node, next.next);
        }
        return new ConcurrentSkipListMap.ValueSpliterator(comparator, head, next, null, (next == null) ? 0 : Integer.MAX_VALUE);
    }
    
    final ConcurrentSkipListMap.EntrySpliterator<K, V> entrySpliterator() {
        Comparator<Object> comparator = this.comparator;
        ConcurrentSkipListMap.HeadIndex<Object, Object> head;
        ConcurrentSkipListMap.Node<Object, Object> next;
        while (true) {
            ConcurrentSkipListMap.Node<K, V> node = (head = this.head).node;
            if ((next = node.next) == null || next.value != null) {
                break;
            }
            next.helpDelete(node, next.next);
        }
        return new ConcurrentSkipListMap.EntrySpliterator(comparator, head, next, null, (next == null) ? 0 : Integer.MAX_VALUE);
    }
    
    static {
        BASE_HEADER = new Object();
        try {
            UNSAFE = Unsafe.getUnsafe();
            headOffset = ConcurrentSkipListMap.UNSAFE.objectFieldOffset(ConcurrentSkipListMap.class.getDeclaredField("head"));
            SECONDARY = ConcurrentSkipListMap.UNSAFE.objectFieldOffset(Thread.class.getDeclaredField("threadLocalRandomSecondarySeed"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
    
    static final class Node<K, V>
    {
        final K key;
        volatile Object value;
        volatile ConcurrentSkipListMap.Node<K, V> next;
        private static final Unsafe UNSAFE;
        private static final long valueOffset;
        private static final long nextOffset;
        
        Node(K key, Object value, ConcurrentSkipListMap.Node<K, V> next) {
            this.key = key;
            this.value = value;
            this.next = next;
        }
        
        Node(ConcurrentSkipListMap.Node<K, V> next) {
            this.key = null;
            this.value = this;
            this.next = next;
        }
        
        boolean casValue(Object expected, Object x) {
            return ConcurrentSkipListMap.Node.UNSAFE.compareAndSwapObject(this, ConcurrentSkipListMap.Node.valueOffset, expected, x);
        }
        
        boolean casNext(ConcurrentSkipListMap.Node<K, V> expected, ConcurrentSkipListMap.Node<K, V> x) {
            return ConcurrentSkipListMap.Node.UNSAFE.compareAndSwapObject(this, ConcurrentSkipListMap.Node.nextOffset, expected, x);
        }
        
        boolean isMarker() {
            return this.value == this;
        }
        
        boolean isBaseHeader() {
            return this.value == ConcurrentSkipListMap.BASE_HEADER;
        }
        
        boolean appendMarker(ConcurrentSkipListMap.Node<K, V> node) {
            return this.casNext(node, new Node<K, V>(node));
        }
        
        void helpDelete(ConcurrentSkipListMap.Node<K, V> node, ConcurrentSkipListMap.Node<K, V> node2) {
            if (node2 == this.next && this == node.next) {
                if (node2 == null || node2.value != node2) {
                    this.casNext(node2, new Node<K, V>(node2));
                }
                else {
                    node.casNext(this, node2.next);
                }
            }
        }
        
        V getValidValue() {
            Object value = this.value;
            if (value == this || value == ConcurrentSkipListMap.BASE_HEADER) {
                return null;
            }
            return value;
        }
        
        AbstractMap.SimpleImmutableEntry<K, V> createSnapshot() {
            Object value = this.value;
            if (value == null || value == this || value == ConcurrentSkipListMap.BASE_HEADER) {
                return null;
            }
            return new AbstractMap.SimpleImmutableEntry(this.key, value);
        }
        
        static {
            try {
                UNSAFE = Unsafe.getUnsafe();
                Class<Object> clazz = ConcurrentSkipListMap.Node.class;
                valueOffset = ConcurrentSkipListMap.Node.UNSAFE.objectFieldOffset(clazz.getDeclaredField("value"));
                nextOffset = ConcurrentSkipListMap.Node.UNSAFE.objectFieldOffset(clazz.getDeclaredField("next"));
            }
            catch (Exception cause) {
                throw new Error(cause);
            }
        }
    }
    
    static class Index<K, V>
    {
        final ConcurrentSkipListMap.Node<K, V> node;
        final ConcurrentSkipListMap.Index<K, V> down;
        volatile ConcurrentSkipListMap.Index<K, V> right;
        private static final Unsafe UNSAFE;
        private static final long rightOffset;
        
        Index(ConcurrentSkipListMap.Node<K, V> node, ConcurrentSkipListMap.Index<K, V> down, ConcurrentSkipListMap.Index<K, V> right) {
            this.node = node;
            this.down = down;
            this.right = right;
        }
        
        final boolean casRight(ConcurrentSkipListMap.Index<K, V> expected, ConcurrentSkipListMap.Index<K, V> x) {
            return ConcurrentSkipListMap.Index.UNSAFE.compareAndSwapObject(this, ConcurrentSkipListMap.Index.rightOffset, expected, x);
        }
        
        final boolean indexesDeletedNode() {
            return this.node.value == null;
        }
        
        final boolean link(ConcurrentSkipListMap.Index<K, V> right, ConcurrentSkipListMap.Index<K, V> index) {
            ConcurrentSkipListMap.Node<Object, Object> node = this.node;
            index.right = right;
            return node.value != null && this.casRight(right, index);
        }
        
        final boolean unlink(ConcurrentSkipListMap.Index<K, V> index) {
            return this.node.value != null && this.casRight(index, index.right);
        }
        
        static {
            try {
                UNSAFE = Unsafe.getUnsafe();
                rightOffset = ConcurrentSkipListMap.Index.UNSAFE.objectFieldOffset(ConcurrentSkipListMap.Index.class.getDeclaredField("right"));
            }
            catch (Exception cause) {
                throw new Error(cause);
            }
        }
    }
    
    static final class HeadIndex<K, V> extends ConcurrentSkipListMap.Index<K, V>
    {
        final int level;
        
        HeadIndex(ConcurrentSkipListMap.Node<K, V> node, ConcurrentSkipListMap.Index<K, V> index, ConcurrentSkipListMap.Index<K, V> index2, int level) {
            super(node, index, index2);
            this.level = level;
        }
    }
    
    abstract class Iter<T> implements Iterator<T>
    {
        ConcurrentSkipListMap.Node<K, V> lastReturned;
        ConcurrentSkipListMap.Node<K, V> next;
        V nextValue;
        
        Iter() {
            while (true) {
                ConcurrentSkipListMap.Node first = ConcurrentSkipListMap.this.findFirst();
                this.next = first;
                if (first != null) {
                    Object value = this.next.value;
                    if (value != null && value != this.next) {
                        this.nextValue = value;
                        break;
                    }
                    continue;
                }
                else {
                    break;
                }
            }
        }
        
        @Override
        public final boolean hasNext() {
            return this.next != null;
        }
        
        final void advance() {
            if (this.next == null) {
                throw new NoSuchElementException();
            }
            this.lastReturned = this.next;
            while (true) {
                ConcurrentSkipListMap.Node<K, V> next = this.next.next;
                this.next = next;
                if (next != null) {
                    Object value = this.next.value;
                    if (value != null && value != this.next) {
                        this.nextValue = value;
                        break;
                    }
                    continue;
                }
                else {
                    break;
                }
            }
            return;
        }
        
        @Override
        public void remove() {
            ConcurrentSkipListMap.Node<Object, Object> lastReturned = this.lastReturned;
            if (lastReturned == null) {
                throw new IllegalStateException();
            }
            ConcurrentSkipListMap.this.remove(lastReturned.key);
            this.lastReturned = null;
            return;
        }
    }
    
    final class ValueIterator extends ConcurrentSkipListMap.Iter<V>
    {
        @Override
        public V next() {
            Object nextValue = this.nextValue;
            this.advance();
            return nextValue;
        }
    }
    
    final class KeyIterator extends ConcurrentSkipListMap.Iter<K>
    {
        @Override
        public K next() {
            ConcurrentSkipListMap.Node<Object, Object> next = this.next;
            this.advance();
            return next.key;
        }
    }
    
    final class EntryIterator extends ConcurrentSkipListMap.Iter<Map.Entry<K, V>>
    {
        @Override
        public Map.Entry<K, V> next() {
            ConcurrentSkipListMap.Node<Object, Object> next = this.next;
            Object nextValue = this.nextValue;
            this.advance();
            return new AbstractMap.SimpleImmutableEntry(next.key, nextValue);
        }
    }
    
    static final class KeySet<E> extends AbstractSet<E> implements NavigableSet<E>
    {
        final ConcurrentNavigableMap<E, ?> m;
        
        KeySet(ConcurrentNavigableMap<E, ?> m) {
            this.m = m;
        }
        
        @Override
        public int size() {
            return this.m.size();
        }
        
        @Override
        public boolean isEmpty() {
            return this.m.isEmpty();
        }
        
        @Override
        public boolean contains(Object o) {
            return this.m.containsKey(o);
        }
        
        @Override
        public boolean remove(Object o) {
            return this.m.remove(o) != null;
        }
        
        @Override
        public void clear() {
            this.m.clear();
        }
        
        @Override
        public E lower(E e) {
            return this.m.lowerKey(e);
        }
        
        @Override
        public E floor(E e) {
            return this.m.floorKey(e);
        }
        
        @Override
        public E ceiling(E e) {
            return this.m.ceilingKey(e);
        }
        
        @Override
        public E higher(E e) {
            return this.m.higherKey(e);
        }
        
        @Override
        public Comparator<? super E> comparator() {
            return this.m.comparator();
        }
        
        @Override
        public E first() {
            return this.m.firstKey();
        }
        
        @Override
        public E last() {
            return this.m.lastKey();
        }
        
        @Override
        public E pollFirst() {
            Map.Entry<Object, Object> pollFirstEntry = this.m.pollFirstEntry();
            return (pollFirstEntry == null) ? null : pollFirstEntry.getKey();
        }
        
        @Override
        public E pollLast() {
            Map.Entry<Object, Object> pollLastEntry = this.m.pollLastEntry();
            return (pollLastEntry == null) ? null : pollLastEntry.getKey();
        }
        
        @Override
        public Iterator<E> iterator() {
            if ((this.m instanceof ConcurrentSkipListMap)) {
                return ((ConcurrentSkipListMap)this.m).keyIterator();
            }
            return ((ConcurrentSkipListMap.SubMap)this.m).keyIterator();
        }
        
        @Override
        public boolean equals(Object o) {
            if (o == this) {
                return true;
            }
            if (!(o instanceof Set)) {
                return false;
            }
            Collection<Object> c = (Collection)o;
            try {
                return this.containsAll(c) && c.containsAll(this);
            }
            catch (ClassCastException ex) {
                return false;
            }
            catch (NullPointerException ex2) {
                return false;
            }
        }
        
        @Override
        public Object[] toArray() {
            return ConcurrentSkipListMap.toList(this).toArray();
        }
        
        @Override
        public <T> T[] toArray(T[] array) {
            return ConcurrentSkipListMap.toList(this).toArray(array);
        }
        
        @Override
        public Iterator<E> descendingIterator() {
            return this.descendingSet().iterator();
        }
        
        @Override
        public NavigableSet<E> subSet(E e, boolean b, E e2, boolean b2) {
            return new ConcurrentSkipListMap.KeySet(this.m.subMap(e, b, e2, b2));
        }
        
        @Override
        public NavigableSet<E> headSet(E e, boolean b) {
            return new ConcurrentSkipListMap.KeySet(this.m.headMap(e, b));
        }
        
        @Override
        public NavigableSet<E> tailSet(E e, boolean b) {
            return new ConcurrentSkipListMap.KeySet(this.m.tailMap(e, b));
        }
        
        @Override
        public NavigableSet<E> subSet(E e, E e2) {
            return this.subSet(e, true, e2, false);
        }
        
        @Override
        public NavigableSet<E> headSet(E e) {
            return this.headSet(e, false);
        }
        
        @Override
        public NavigableSet<E> tailSet(E e) {
            return this.tailSet(e, true);
        }
        
        @Override
        public NavigableSet<E> descendingSet() {
            return new ConcurrentSkipListMap.KeySet(this.m.descendingMap());
        }
        
        @Override
        public Spliterator<E> spliterator() {
            if ((this.m instanceof ConcurrentSkipListMap)) {
                return ((ConcurrentSkipListMap)this.m).keySpliterator();
            }
            return (Spliterator)((ConcurrentSkipListMap.SubMap)this.m).keyIterator();
        }
    }
    
    static final class Values<E> extends AbstractCollection<E>
    {
        final ConcurrentNavigableMap<?, E> m;
        
        Values(ConcurrentNavigableMap<?, E> m) {
            this.m = m;
        }
        
        @Override
        public Iterator<E> iterator() {
            if ((this.m instanceof ConcurrentSkipListMap)) {
                return ((ConcurrentSkipListMap)this.m).valueIterator();
            }
            return ((ConcurrentSkipListMap.SubMap)this.m).valueIterator();
        }
        
        @Override
        public boolean isEmpty() {
            return this.m.isEmpty();
        }
        
        @Override
        public int size() {
            return this.m.size();
        }
        
        @Override
        public boolean contains(Object o) {
            return this.m.containsValue(o);
        }
        
        @Override
        public void clear() {
            this.m.clear();
        }
        
        @Override
        public Object[] toArray() {
            return ConcurrentSkipListMap.toList(this).toArray();
        }
        
        @Override
        public <T> T[] toArray(T[] array) {
            return ConcurrentSkipListMap.toList(this).toArray(array);
        }
        
        @Override
        public Spliterator<E> spliterator() {
            if ((this.m instanceof ConcurrentSkipListMap)) {
                return ((ConcurrentSkipListMap)this.m).valueSpliterator();
            }
            return (Spliterator)((ConcurrentSkipListMap.SubMap)this.m).valueIterator();
        }
    }
    
    static final class EntrySet<K1, V1> extends AbstractSet<Map.Entry<K1, V1>>
    {
        final ConcurrentNavigableMap<K1, V1> m;
        
        EntrySet(ConcurrentNavigableMap<K1, V1> m) {
            this.m = m;
        }
        
        @Override
        public Iterator<Map.Entry<K1, V1>> iterator() {
            if ((this.m instanceof ConcurrentSkipListMap)) {
                return ((ConcurrentSkipListMap)this.m).entryIterator();
            }
            return ((ConcurrentSkipListMap.SubMap)this.m).entryIterator();
        }
        
        @Override
        public boolean contains(Object o) {
            if (!(o instanceof Map.Entry)) {
                return false;
            }
            Map.Entry<Object, Object> entry = (Map.Entry)o;
            Object value = this.m.get(entry.getKey());
            return value != null && value.equals(entry.getValue());
        }
        
        @Override
        public boolean remove(Object o) {
            if (!(o instanceof Map.Entry)) {
                return false;
            }
            Map.Entry<Object, Object> entry = (Map.Entry)o;
            return this.m.remove(entry.getKey(), entry.getValue());
        }
        
        @Override
        public boolean isEmpty() {
            return this.m.isEmpty();
        }
        
        @Override
        public int size() {
            return this.m.size();
        }
        
        @Override
        public void clear() {
            this.m.clear();
        }
        
        @Override
        public boolean equals(Object o) {
            if (o == this) {
                return true;
            }
            if (!(o instanceof Set)) {
                return false;
            }
            Collection<Object> c = (Collection)o;
            try {
                return this.containsAll(c) && c.containsAll(this);
            }
            catch (ClassCastException ex) {
                return false;
            }
            catch (NullPointerException ex2) {
                return false;
            }
        }
        
        @Override
        public Object[] toArray() {
            return ConcurrentSkipListMap.toList(this).toArray();
        }
        
        @Override
        public <T> T[] toArray(T[] array) {
            return ConcurrentSkipListMap.toList(this).toArray(array);
        }
        
        @Override
        public Spliterator<Map.Entry<K1, V1>> spliterator() {
            if ((this.m instanceof ConcurrentSkipListMap)) {
                return ((ConcurrentSkipListMap)this.m).entrySpliterator();
            }
            return (Spliterator)((ConcurrentSkipListMap.SubMap)this.m).entryIterator();
        }
    }
    
    static final class SubMap<K, V> extends AbstractMap<K, V> implements ConcurrentNavigableMap<K, V>, Cloneable, Serializable
    {
        private static final long serialVersionUID = -7647078645895051609L;
        private final ConcurrentSkipListMap<K, V> m;
        private final K lo;
        private final K hi;
        private final boolean loInclusive;
        private final boolean hiInclusive;
        private final boolean isDescending;
        private transient ConcurrentSkipListMap.KeySet<K> keySetView;
        private transient Set<Map.Entry<K, V>> entrySetView;
        private transient Collection<V> valuesView;
        
        SubMap(ConcurrentSkipListMap<K, V> m, K lo, boolean loInclusive, K hi, boolean hiInclusive, boolean isDescending) {
            Comparator<Object> comparator = m.comparator;
            if (lo != null && hi != null && ConcurrentSkipListMap.cpr(comparator, lo, hi) > 0) {
                throw new IllegalArgumentException("inconsistent range");
            }
            this.m = m;
            this.lo = lo;
            this.hi = hi;
            this.loInclusive = loInclusive;
            this.hiInclusive = hiInclusive;
            this.isDescending = isDescending;
            return;
        }
        
        boolean tooLow(Object o, Comparator<? super K> comparator) {
            int cpr;
            return this.lo != null && ((cpr = ConcurrentSkipListMap.cpr(comparator, o, this.lo)) < 0 || (cpr == 0 && !this.loInclusive));
        }
        
        boolean tooHigh(Object o, Comparator<? super K> comparator) {
            int cpr;
            return this.hi != null && ((cpr = ConcurrentSkipListMap.cpr(comparator, o, this.hi)) > 0 || (cpr == 0 && !this.hiInclusive));
        }
        
        boolean inBounds(Object o, Comparator<? super K> comparator) {
            return !this.tooLow(o, comparator) && !this.tooHigh(o, comparator);
        }
        
        void checkKeyBounds(K k, Comparator<? super K> comparator) {
            if (k == null) {
                throw new NullPointerException();
            }
            if (!this.inBounds(k, comparator)) {
                throw new IllegalArgumentException("key out of range");
            }
            return;
        }
        
        boolean isBeforeEnd(ConcurrentSkipListMap.Node<K, V> node, Comparator<? super K> comparator) {
            if (node == null) {
                return false;
            }
            if (this.hi == null) {
                return true;
            }
            Object key = node.key;
            if (key == null) {
                return true;
            }
            int cpr = ConcurrentSkipListMap.cpr(comparator, key, this.hi);
            return cpr <= 0 && (cpr != 0 || this.hiInclusive);
        }
        
        ConcurrentSkipListMap.Node<K, V> loNode(Comparator<? super K> comparator) {
            if (this.lo == null) {
                return this.m.findFirst();
            }
            if (this.loInclusive) {
                return this.m.findNear(this.lo, 1, comparator);
            }
            return this.m.findNear(this.lo, 0, comparator);
        }
        
        ConcurrentSkipListMap.Node<K, V> hiNode(Comparator<? super K> comparator) {
            if (this.hi == null) {
                return this.m.findLast();
            }
            if (this.hiInclusive) {
                return this.m.findNear(this.hi, 3, comparator);
            }
            return this.m.findNear(this.hi, 2, comparator);
        }
        
        K lowestKey() {
            Comparator<Object> comparator = this.m.comparator;
            ConcurrentSkipListMap.Node<K, V> loNode = this.loNode(comparator);
            if (this.isBeforeEnd(loNode, comparator)) {
                return loNode.key;
            }
            throw new NoSuchElementException();
        }
        
        K highestKey() {
            Comparator<Object> comparator = this.m.comparator;
            ConcurrentSkipListMap.Node<K, V> hiNode = this.hiNode(comparator);
            if (hiNode != null) {
                Object key = hiNode.key;
                if (this.inBounds(key, comparator)) {
                    return key;
                }
            }
            throw new NoSuchElementException();
        }
        
        Map.Entry<K, V> lowestEntry() {
            Comparator<Object> comparator = this.m.comparator;
            while (true) {
                ConcurrentSkipListMap.Node<K, V> loNode = this.loNode(comparator);
                if (!this.isBeforeEnd(loNode, comparator)) {
                    return null;
                }
                AbstractMap.SimpleImmutableEntry<Object, Object> snapshot = loNode.createSnapshot();
                if (snapshot != null) {
                    return snapshot;
                }
                continue;
            }
        }
        
        Map.Entry<K, V> highestEntry() {
            Comparator<Object> comparator = this.m.comparator;
            while (true) {
                ConcurrentSkipListMap.Node<K, V> hiNode = this.hiNode(comparator);
                if (hiNode == null || !this.inBounds(hiNode.key, comparator)) {
                    return null;
                }
                AbstractMap.SimpleImmutableEntry<Object, Object> snapshot = hiNode.createSnapshot();
                if (snapshot != null) {
                    return snapshot;
                }
                continue;
            }
        }
        
        Map.Entry<K, V> removeLowest() {
            Comparator<Object> comparator = this.m.comparator;
            while (true) {
                ConcurrentSkipListMap.Node<K, V> loNode = this.loNode(comparator);
                if (loNode == null) {
                    return null;
                }
                Object key = loNode.key;
                if (!this.inBounds(key, comparator)) {
                    return null;
                }
                Object doRemove = this.m.doRemove(key, null);
                if (doRemove != null) {
                    return new AbstractMap.SimpleImmutableEntry(key, doRemove);
                }
                continue;
            }
        }
        
        Map.Entry<K, V> removeHighest() {
            Comparator<Object> comparator = this.m.comparator;
            while (true) {
                ConcurrentSkipListMap.Node<K, V> hiNode = this.hiNode(comparator);
                if (hiNode == null) {
                    return null;
                }
                Object key = hiNode.key;
                if (!this.inBounds(key, comparator)) {
                    return null;
                }
                Object doRemove = this.m.doRemove(key, null);
                if (doRemove != null) {
                    return new AbstractMap.SimpleImmutableEntry(key, doRemove);
                }
                continue;
            }
        }
        
        Map.Entry<K, V> getNearEntry(K k, int n) {
            Comparator<Object> comparator = this.m.comparator;
            if (this.isDescending) {
                if ((n & 0x2) == 0x0) {
                    n |= 0x2;
                }
                else {
                    n &= 0xFFFFFFFD;
                }
            }
            if (this.tooLow(k, comparator)) {
                return ((n & 0x2) != 0x0) ? null : this.lowestEntry();
            }
            if (this.tooHigh(k, comparator)) {
                return ((n & 0x2) != 0x0) ? this.highestEntry() : null;
            }
            while (true) {
                ConcurrentSkipListMap.Node<K, V> near = this.m.findNear(k, n, comparator);
                if (near == null || !this.inBounds(near.key, comparator)) {
                    return null;
                }
                Object key = near.key;
                Object validValue = near.getValidValue();
                if (validValue != null) {
                    return new AbstractMap.SimpleImmutableEntry(key, validValue);
                }
                continue;
            }
        }
        
        K getNearKey(K k, int n) {
            Comparator<Object> comparator = this.m.comparator;
            if (this.isDescending) {
                if ((n & 0x2) == 0x0) {
                    n |= 0x2;
                }
                else {
                    n &= 0xFFFFFFFD;
                }
            }
            if (this.tooLow(k, comparator)) {
                if ((n & 0x2) == 0x0) {
                    ConcurrentSkipListMap.Node<K, V> loNode = this.loNode(comparator);
                    if (this.isBeforeEnd(loNode, comparator)) {
                        return loNode.key;
                    }
                }
                return null;
            }
            else if (this.tooHigh(k, comparator)) {
                if ((n & 0x2) != 0x0) {
                    ConcurrentSkipListMap.Node<K, V> hiNode = this.hiNode(comparator);
                    if (hiNode != null) {
                        Object key2 = hiNode.key;
                        if (this.inBounds(key2, comparator)) {
                            return key2;
                        }
                    }
                }
                return null;
            }
            else {
                while (true) {
                    ConcurrentSkipListMap.Node<K, V> near = this.m.findNear(k, n, comparator);
                    if (near == null || !this.inBounds(near.key, comparator)) {
                        return null;
                    }
                    Object key = near.key;
                    if (near.getValidValue() != null) {
                        return key;
                    }
                    continue;
                }
            }
        }
        
        @Override
        public boolean containsKey(Object o) {
            if (o == null) {
                throw new NullPointerException();
            }
            return this.inBounds(o, this.m.comparator) && this.m.containsKey(o);
        }
        
        @Override
        public V get(Object o) {
            if (o == null) {
                throw new NullPointerException();
            }
            return this.inBounds(o, this.m.comparator) ? this.m.get(o) : null;
        }
        
        @Override
        public V put(K k, V v) {
            this.checkKeyBounds(k, this.m.comparator);
            return this.m.put(k, v);
        }
        
        @Override
        public V remove(Object o) {
            return this.inBounds(o, this.m.comparator) ? this.m.remove(o) : null;
        }
        
        @Override
        public int size() {
            Comparator<Object> comparator = this.m.comparator;
            long n = 0L;
            for (ConcurrentSkipListMap.Node node = this.loNode(comparator); this.isBeforeEnd(node, comparator); node = node.next) {
                if (node.getValidValue() != null) {
                    ++n;
                }
            }
            return (n >= 2147483647L) ? Integer.MAX_VALUE : ((int)n);
        }
        
        @Override
        public boolean isEmpty() {
            Comparator<Object> comparator = this.m.comparator;
            return !this.isBeforeEnd(this.loNode(comparator), comparator);
        }
        
        @Override
        public boolean containsValue(Object o) {
            if (o == null) {
                throw new NullPointerException();
            }
            Comparator<Object> comparator = this.m.comparator;
            ConcurrentSkipListMap.Node<K, V> node = this.loNode(comparator);
            while (this.isBeforeEnd(node, comparator)) {
                Object validValue = node.getValidValue();
                if (validValue != null && o.equals(validValue)) {
                    return true;
                }
                node = node.next;
            }
            return false;
        }
        
        @Override
        public void clear() {
            Comparator<Object> comparator = this.m.comparator;
            for (ConcurrentSkipListMap.Node node = this.loNode(comparator); this.isBeforeEnd(node, comparator); node = node.next) {
                if (node.getValidValue() != null) {
                    this.m.remove(node.key);
                }
            }
        }
        
        @Override
        public V putIfAbsent(K k, V v) {
            this.checkKeyBounds(k, this.m.comparator);
            return this.m.putIfAbsent(k, v);
        }
        
        @Override
        public boolean remove(Object o, Object o2) {
            return this.inBounds(o, this.m.comparator) && this.m.remove(o, o2);
        }
        
        @Override
        public boolean replace(K k, V v, V v2) {
            this.checkKeyBounds(k, this.m.comparator);
            return this.m.replace(k, v, v2);
        }
        
        @Override
        public V replace(K k, V v) {
            this.checkKeyBounds(k, this.m.comparator);
            return this.m.replace(k, v);
        }
        
        @Override
        public Comparator<? super K> comparator() {
            Comparator<Object> comparator = this.m.comparator();
            if (this.isDescending) {
                return Collections.reverseOrder(comparator);
            }
            return comparator;
        }
        
        ConcurrentSkipListMap.SubMap<K, V> newSubMap(K lo, boolean b, K hi, boolean hiInclusive) {
            Comparator<Object> comparator = this.m.comparator;
            if (this.isDescending) {
                lo = (hi = hi);
                loInclusive = (hiInclusive = hiInclusive);
            }
            if (this.lo != null) {
                if (lo == null) {
                    lo = this.lo;
                    loInclusive = this.loInclusive;
                }
                else {
                    int cpr = ConcurrentSkipListMap.cpr(comparator, lo, this.lo);
                    if (cpr < 0 || (cpr == 0 && !this.loInclusive && loInclusive)) {
                        throw new IllegalArgumentException("key out of range");
                    }
                }
            }
            if (this.hi != null) {
                if (hi == null) {
                    hi = this.hi;
                    hiInclusive = this.hiInclusive;
                }
                else {
                    int cpr2 = ConcurrentSkipListMap.cpr(comparator, hi, this.hi);
                    if (cpr2 > 0 || (cpr2 == 0 && !this.hiInclusive && hiInclusive)) {
                        throw new IllegalArgumentException("key out of range");
                    }
                }
            }
            return new ConcurrentSkipListMap.SubMap<K, V>(this.m, lo, loInclusive, hi, hiInclusive, this.isDescending);
        }
        
        @Override
        public ConcurrentSkipListMap.SubMap<K, V> subMap(K k, boolean b, K i, boolean b2) {
            if (k == null || i == null) {
                throw new NullPointerException();
            }
            return this.newSubMap(k, b, i, b2);
        }
        
        @Override
        public ConcurrentSkipListMap.SubMap<K, V> headMap(K k, boolean b) {
            if (k == null) {
                throw new NullPointerException();
            }
            return this.newSubMap(null, false, k, b);
        }
        
        @Override
        public ConcurrentSkipListMap.SubMap<K, V> tailMap(K k, boolean b) {
            if (k == null) {
                throw new NullPointerException();
            }
            return this.newSubMap(k, b, null, false);
        }
        
        @Override
        public ConcurrentSkipListMap.SubMap<K, V> subMap(K k, K i) {
            return this.subMap(k, true, i, false);
        }
        
        @Override
        public ConcurrentSkipListMap.SubMap<K, V> headMap(K k) {
            return this.headMap(k, false);
        }
        
        @Override
        public ConcurrentSkipListMap.SubMap<K, V> tailMap(K k) {
            return this.tailMap(k, true);
        }
        
        @Override
        public ConcurrentSkipListMap.SubMap<K, V> descendingMap() {
            return new ConcurrentSkipListMap.SubMap<K, V>(this.m, this.lo, this.loInclusive, this.hi, this.hiInclusive, !this.isDescending);
        }
        
        @Override
        public Map.Entry<K, V> ceilingEntry(K k) {
            return this.getNearEntry(k, 1);
        }
        
        @Override
        public K ceilingKey(K k) {
            return this.getNearKey(k, 1);
        }
        
        @Override
        public Map.Entry<K, V> lowerEntry(K k) {
            return this.getNearEntry(k, 2);
        }
        
        @Override
        public K lowerKey(K k) {
            return this.getNearKey(k, 2);
        }
        
        @Override
        public Map.Entry<K, V> floorEntry(K k) {
            return this.getNearEntry(k, 3);
        }
        
        @Override
        public K floorKey(K k) {
            return this.getNearKey(k, 3);
        }
        
        @Override
        public Map.Entry<K, V> higherEntry(K k) {
            return this.getNearEntry(k, 0);
        }
        
        @Override
        public K higherKey(K k) {
            return this.getNearKey(k, 0);
        }
        
        @Override
        public K firstKey() {
            return this.isDescending ? this.highestKey() : this.lowestKey();
        }
        
        @Override
        public K lastKey() {
            return this.isDescending ? this.lowestKey() : this.highestKey();
        }
        
        @Override
        public Map.Entry<K, V> firstEntry() {
            return this.isDescending ? this.highestEntry() : this.lowestEntry();
        }
        
        @Override
        public Map.Entry<K, V> lastEntry() {
            return this.isDescending ? this.lowestEntry() : this.highestEntry();
        }
        
        @Override
        public Map.Entry<K, V> pollFirstEntry() {
            return this.isDescending ? this.removeHighest() : this.removeLowest();
        }
        
        @Override
        public Map.Entry<K, V> pollLastEntry() {
            return this.isDescending ? this.removeLowest() : this.removeHighest();
        }
        
        @Override
        public NavigableSet<K> keySet() {
            ConcurrentSkipListMap.KeySet<Object> keySetView = this.keySetView;
            return (keySetView != null) ? keySetView : (this.keySetView = new ConcurrentSkipListMap.KeySet(this));
        }
        
        @Override
        public NavigableSet<K> navigableKeySet() {
            ConcurrentSkipListMap.KeySet<Object> keySetView = this.keySetView;
            return (keySetView != null) ? keySetView : (this.keySetView = new ConcurrentSkipListMap.KeySet(this));
        }
        
        @Override
        public Collection<V> values() {
            Collection<Object> valuesView = this.valuesView;
            return (valuesView != null) ? valuesView : (this.valuesView = new ConcurrentSkipListMap.Values(this));
        }
        
        @Override
        public Set<Map.Entry<K, V>> entrySet() {
            Set<Object> entrySetView = this.entrySetView;
            return (entrySetView != null) ? entrySetView : (this.entrySetView = new ConcurrentSkipListMap.EntrySet(this));
        }
        
        @Override
        public NavigableSet<K> descendingKeySet() {
            return this.descendingMap().navigableKeySet();
        }
        
        Iterator<K> keyIterator() {
            return new ConcurrentSkipListMap.SubMap.SubMapKeyIterator();
        }
        
        Iterator<V> valueIterator() {
            return new ConcurrentSkipListMap.SubMap.SubMapValueIterator();
        }
        
        Iterator<Map.Entry<K, V>> entryIterator() {
            return new ConcurrentSkipListMap.SubMap.SubMapEntryIterator();
        }
        
        abstract class SubMapIter<T> implements Iterator<T>, Spliterator<T>
        {
            ConcurrentSkipListMap.Node<K, V> lastReturned;
            ConcurrentSkipListMap.Node<K, V> next;
            V nextValue;
            
            SubMapIter() {
                Comparator<Object> comparator = ConcurrentSkipListMap.SubMap.this.m.comparator;
                while (true) {
                    this.next = (ConcurrentSkipListMap.SubMap.this.isDescending ? ConcurrentSkipListMap.SubMap.this.hiNode(comparator) : ConcurrentSkipListMap.SubMap.this.loNode(comparator));
                    if (this.next == null) {
                        break;
                    }
                    Object value = this.next.value;
                    if (value != null && value != this.next) {
                        if (!ConcurrentSkipListMap.SubMap.this.inBounds(this.next.key, comparator)) {
                            this.next = null;
                            break;
                        }
                        this.nextValue = value;
                        break;
                    }
                    else {
                        continue;
                    }
                }
            }
            
            @Override
            public final boolean hasNext() {
                return this.next != null;
            }
            
            final void advance() {
                if (this.next == null) {
                    throw new NoSuchElementException();
                }
                this.lastReturned = this.next;
                if (ConcurrentSkipListMap.SubMap.this.isDescending) {
                    this.descend();
                }
                else {
                    this.ascend();
                }
                return;
            }
            
            private void ascend() {
                Comparator<Object> comparator = ConcurrentSkipListMap.SubMap.this.m.comparator;
                while (true) {
                    this.next = this.next.next;
                    if (this.next == null) {
                        break;
                    }
                    Object value = this.next.value;
                    if (value != null && value != this.next) {
                        if (ConcurrentSkipListMap.SubMap.this.tooHigh(this.next.key, comparator)) {
                            this.next = null;
                            break;
                        }
                        this.nextValue = value;
                        break;
                    }
                    else {
                        continue;
                    }
                }
            }
            
            private void descend() {
                Comparator<Object> comparator = ConcurrentSkipListMap.SubMap.this.m.comparator;
                while (true) {
                    this.next = ConcurrentSkipListMap.SubMap.this.m.findNear(this.lastReturned.key, 2, comparator);
                    if (this.next == null) {
                        break;
                    }
                    Object value = this.next.value;
                    if (value != null && value != this.next) {
                        if (ConcurrentSkipListMap.SubMap.this.tooLow(this.next.key, comparator)) {
                            this.next = null;
                            break;
                        }
                        this.nextValue = value;
                        break;
                    }
                    else {
                        continue;
                    }
                }
            }
            
            @Override
            public void remove() {
                ConcurrentSkipListMap.Node<Object, Object> lastReturned = this.lastReturned;
                if (lastReturned == null) {
                    throw new IllegalStateException();
                }
                ConcurrentSkipListMap.SubMap.this.m.remove(lastReturned.key);
                this.lastReturned = null;
                return;
            }
            
            @Override
            public Spliterator<T> trySplit() {
                return null;
            }
            
            @Override
            public boolean tryAdvance(Consumer<? super T> consumer) {
                if (this.hasNext()) {
                    consumer.accept(this.next());
                    return true;
                }
                return false;
            }
            
            @Override
            public void forEachRemaining(Consumer<? super T> consumer) {
                while (this.hasNext()) {
                    consumer.accept(this.next());
                }
            }
            
            @Override
            public long estimateSize() {
                return Long.MAX_VALUE;
            }
        }
        
        final class SubMapValueIterator extends ConcurrentSkipListMap.SubMap.SubMapIter<V>
        {
            @Override
            public V next() {
                Object nextValue = this.nextValue;
                this.advance();
                return nextValue;
            }
            
            @Override
            public int characteristics() {
                return 0;
            }
        }
        
        final class SubMapKeyIterator extends ConcurrentSkipListMap.SubMap.SubMapIter<K>
        {
            @Override
            public K next() {
                ConcurrentSkipListMap.Node<Object, Object> next = this.next;
                this.advance();
                return next.key;
            }
            
            @Override
            public int characteristics() {
                return 21;
            }
            
            @Override
            public final Comparator<? super K> getComparator() {
                return ConcurrentSkipListMap.SubMap.this.comparator();
            }
        }
        
        final class SubMapEntryIterator extends ConcurrentSkipListMap.SubMap.SubMapIter<Map.Entry<K, V>>
        {
            @Override
            public Map.Entry<K, V> next() {
                ConcurrentSkipListMap.Node<Object, Object> next = this.next;
                Object nextValue = this.nextValue;
                this.advance();
                return new AbstractMap.SimpleImmutableEntry(next.key, nextValue);
            }
            
            @Override
            public int characteristics() {
                return 1;
            }
        }
    }
    
    abstract static class CSLMSpliterator<K, V>
    {
        final Comparator<? super K> comparator;
        final K fence;
        ConcurrentSkipListMap.Index<K, V> row;
        ConcurrentSkipListMap.Node<K, V> current;
        int est;
        
        CSLMSpliterator(Comparator<? super K> comparator, ConcurrentSkipListMap.Index<K, V> row, ConcurrentSkipListMap.Node<K, V> current, K fence, int est) {
            this.comparator = comparator;
            this.row = row;
            this.current = current;
            this.fence = fence;
            this.est = est;
        }
        
        public final long estimateSize() {
            return (long)this.est;
        }
    }
    
    static final class KeySpliterator<K, V> extends ConcurrentSkipListMap.CSLMSpliterator<K, V> implements Spliterator<K>
    {
        KeySpliterator(Comparator<? super K> comparator, ConcurrentSkipListMap.Index<K, V> index, ConcurrentSkipListMap.Node<K, V> node, K k, int n) {
            super(comparator, index, node, k, n);
        }
        
        @Override
        public Spliterator<K> trySplit() {
            Comparator<Object> comparator = this.comparator;
            Object fence = this.fence;
            ConcurrentSkipListMap.Node<Object, Object> current;
            Object key;
            if ((current = this.current) != null && (key = current.key) != null) {
                ConcurrentSkipListMap.Index<K, V> row = this.row;
                while (row != null) {
                    ConcurrentSkipListMap.Index<Object, Object> right;
                    ConcurrentSkipListMap.Node<Object, Object> node;
                    ConcurrentSkipListMap.Node<Object, Object> next;
                    Object key2;
                    if ((right = row.right) != null && (node = right.node) != null && (next = node.next) != null && next.value != null && (key2 = next.key) != null && ConcurrentSkipListMap.cpr(comparator, key2, key) > 0 && (fence == null || ConcurrentSkipListMap.cpr(comparator, key2, fence) < 0)) {
                        this.current = next;
                        ConcurrentSkipListMap.Index<Object, Object> down2 = row.down;
                        this.row = ((right.right != null) ? right : right.down);
                        this.est -= this.est >>> 2;
                        return new ConcurrentSkipListMap.KeySpliterator(comparator, down2, current, key2, this.est);
                    }
                    ConcurrentSkipListMap.Index<K, V> down = row.down;
                    this.row = down;
                    row = down;
                }
            }
            return null;
        }
        
        @Override
        public void forEachRemaining(Consumer<? super K> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            Comparator<Object> comparator = this.comparator;
            Object fence = this.fence;
            ConcurrentSkipListMap.Node<Object, Object> node = this.current;
            this.current = null;
            Object key;
            while (node != null && ((key = node.key) == null || fence == null || ConcurrentSkipListMap.cpr(comparator, fence, key) > 0)) {
                Object value;
                if ((value = node.value) != null && value != node) {
                    consumer.accept(key);
                }
                node = node.next;
            }
            return;
        }
        
        @Override
        public boolean tryAdvance(Consumer<? super K> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            Comparator<Object> comparator = this.comparator;
            Object fence = this.fence;
            ConcurrentSkipListMap.Node<Object, Object> current = this.current;
            while (current != null) {
                Object key;
                if ((key = current.key) != null && fence != null && ConcurrentSkipListMap.cpr(comparator, fence, key) <= 0) {
                    current = null;
                    break;
                }
                Object value;
                if ((value = current.value) != null && value != current) {
                    this.current = current.next;
                    consumer.accept(key);
                    return true;
                }
                current = current.next;
            }
            this.current = current;
            return false;
        }
        
        @Override
        public int characteristics() {
            return 4373;
        }
        
        @Override
        public final Comparator<? super K> getComparator() {
            return this.comparator;
        }
    }
    
    static final class ValueSpliterator<K, V> extends ConcurrentSkipListMap.CSLMSpliterator<K, V> implements Spliterator<V>
    {
        ValueSpliterator(Comparator<? super K> comparator, ConcurrentSkipListMap.Index<K, V> index, ConcurrentSkipListMap.Node<K, V> node, K k, int n) {
            super(comparator, index, node, k, n);
        }
        
        @Override
        public Spliterator<V> trySplit() {
            Comparator<Object> comparator = this.comparator;
            Object fence = this.fence;
            ConcurrentSkipListMap.Node<Object, Object> current;
            Object key;
            if ((current = this.current) != null && (key = current.key) != null) {
                ConcurrentSkipListMap.Index<K, V> row = this.row;
                while (row != null) {
                    ConcurrentSkipListMap.Index<Object, Object> right;
                    ConcurrentSkipListMap.Node<Object, Object> node;
                    ConcurrentSkipListMap.Node<Object, Object> next;
                    Object key2;
                    if ((right = row.right) != null && (node = right.node) != null && (next = node.next) != null && next.value != null && (key2 = next.key) != null && ConcurrentSkipListMap.cpr(comparator, key2, key) > 0 && (fence == null || ConcurrentSkipListMap.cpr(comparator, key2, fence) < 0)) {
                        this.current = next;
                        ConcurrentSkipListMap.Index<Object, Object> down2 = row.down;
                        this.row = ((right.right != null) ? right : right.down);
                        this.est -= this.est >>> 2;
                        return new ConcurrentSkipListMap.ValueSpliterator(comparator, down2, current, key2, this.est);
                    }
                    ConcurrentSkipListMap.Index<K, V> down = row.down;
                    this.row = down;
                    row = down;
                }
            }
            return null;
        }
        
        @Override
        public void forEachRemaining(Consumer<? super V> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            Comparator<Object> comparator = this.comparator;
            Object fence = this.fence;
            ConcurrentSkipListMap.Node<Object, Object> node = this.current;
            this.current = null;
            Object key;
            while (node != null && ((key = node.key) == null || fence == null || ConcurrentSkipListMap.cpr(comparator, fence, key) > 0)) {
                Object value;
                if ((value = node.value) != null && value != node) {
                    consumer.accept(value);
                }
                node = node.next;
            }
            return;
        }
        
        @Override
        public boolean tryAdvance(Consumer<? super V> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            Comparator<Object> comparator = this.comparator;
            Object fence = this.fence;
            ConcurrentSkipListMap.Node<Object, Object> current = this.current;
            while (current != null) {
                Object key;
                if ((key = current.key) != null && fence != null && ConcurrentSkipListMap.cpr(comparator, fence, key) <= 0) {
                    current = null;
                    break;
                }
                Object value;
                if ((value = current.value) != null && value != current) {
                    this.current = current.next;
                    consumer.accept(value);
                    return true;
                }
                current = current.next;
            }
            this.current = current;
            return false;
        }
        
        @Override
        public int characteristics() {
            return 4368;
        }
    }
    
    static final class EntrySpliterator<K, V> extends ConcurrentSkipListMap.CSLMSpliterator<K, V> implements Spliterator<Map.Entry<K, V>>
    {
        EntrySpliterator(Comparator<? super K> comparator, ConcurrentSkipListMap.Index<K, V> index, ConcurrentSkipListMap.Node<K, V> node, K k, int n) {
            super(comparator, index, node, k, n);
        }
        
        @Override
        public Spliterator<Map.Entry<K, V>> trySplit() {
            Comparator<Object> comparator = this.comparator;
            Object fence = this.fence;
            ConcurrentSkipListMap.Node<Object, Object> current;
            Object key;
            if ((current = this.current) != null && (key = current.key) != null) {
                ConcurrentSkipListMap.Index<K, V> row = this.row;
                while (row != null) {
                    ConcurrentSkipListMap.Index<Object, Object> right;
                    ConcurrentSkipListMap.Node<Object, Object> node;
                    ConcurrentSkipListMap.Node<Object, Object> next;
                    Object key2;
                    if ((right = row.right) != null && (node = right.node) != null && (next = node.next) != null && next.value != null && (key2 = next.key) != null && ConcurrentSkipListMap.cpr(comparator, key2, key) > 0 && (fence == null || ConcurrentSkipListMap.cpr(comparator, key2, fence) < 0)) {
                        this.current = next;
                        ConcurrentSkipListMap.Index<Object, Object> down2 = row.down;
                        this.row = ((right.right != null) ? right : right.down);
                        this.est -= this.est >>> 2;
                        return new ConcurrentSkipListMap.EntrySpliterator(comparator, down2, current, key2, this.est);
                    }
                    ConcurrentSkipListMap.Index<K, V> down = row.down;
                    this.row = down;
                    row = down;
                }
            }
            return null;
        }
        
        @Override
        public void forEachRemaining(Consumer<? super Map.Entry<K, V>> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            Comparator<Object> comparator = this.comparator;
            Object fence = this.fence;
            ConcurrentSkipListMap.Node<Object, Object> node = this.current;
            this.current = null;
            Object key;
            while (node != null && ((key = node.key) == null || fence == null || ConcurrentSkipListMap.cpr(comparator, fence, key) > 0)) {
                Object value;
                if ((value = node.value) != null && value != node) {
                    consumer.accept(new AbstractMap.SimpleImmutableEntry(key, value));
                }
                node = node.next;
            }
            return;
        }
        
        @Override
        public boolean tryAdvance(Consumer<? super Map.Entry<K, V>> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            Comparator<Object> comparator = this.comparator;
            Object fence = this.fence;
            ConcurrentSkipListMap.Node<Object, Object> current = this.current;
            while (current != null) {
                Object key;
                if ((key = current.key) != null && fence != null && ConcurrentSkipListMap.cpr(comparator, fence, key) <= 0) {
                    current = null;
                    break;
                }
                Object value;
                if ((value = current.value) != null && value != current) {
                    this.current = current.next;
                    consumer.accept(new AbstractMap.SimpleImmutableEntry(key, value));
                    return true;
                }
                current = current.next;
            }
            this.current = current;
            return false;
        }
        
        @Override
        public int characteristics() {
            return 4373;
        }
        
        @Override
        public final Comparator<Map.Entry<K, V>> getComparator() {
            if (this.comparator != null) {
                return Map.Entry.comparingByKey(this.comparator);
            }
            return (Comparator)(Serializable)((entry, entry2)/* java.util.Map.Entry,java.util.Map.Entry, */ -> ((Comparable)entry.getKey()).compareTo(entry2.getKey()));
        }
    }
}
